//=============================================================================
//
//	control.cpp
//	2010.01.20 Tada 	  
//
//=============================================================================

#include "control.h"
#include "cvtogl.h"

//ラベリング
//#include "Labeling.h"
LabelingBS labeling;
using namespace std;

Bumblebee2Control::Bumblebee2Control(){

}

Bumblebee2Control::~Bumblebee2Control(){
	//closeB();
}

int Bumblebee2Control::getWidth(){
	return stereoCamera.nCols;
}

int Bumblebee2Control::getHeight(){
	return stereoCamera.nRows;	
}

int Bumblebee2Control::getChannels(){
    return stereoCamera.nBytesPerPixel;	
}

int Bumblebee2Control::getColor(){
	return stereoCamera.bColor;
}


int Bumblebee2Control::closeB(){

    printf( "Stop transmission\n" );
    //  Stop data transmission
    if ( dc1394_video_set_transmission( stereoCamera.camera, DC1394_OFF ) != DC1394_SUCCESS )
	{
	    fprintf( stderr, "Couldn't stop the camera?\n" );
	}
 
	delete[] pucDeInterlacedBuffer;
    if ( pucRGBBuffer )	
	delete[] pucRGBBuffer;
    if ( pucGreenBuffer )	
	delete[] pucGreenBuffer;

	// close camera
    cleanup_and_exit( camera );
	return 0;
}


//=============================================================================
// cleanup_and_exit()
// This is called when the program exits and destroys the existing connections
// to the 1394 drivers
void Bumblebee2Control::cleanup_and_exit( dc1394camera_t* camera )
{
	dc1394_capture_stop( camera );		//カメラ関係のハンドルを処理してから終了
	dc1394_video_set_transmission( camera, DC1394_OFF );
	dc1394_camera_free( camera );
	//exit( 0 );
}


int Bumblebee2Control::Initialize(){
	d = dc1394_new ();								//バス全体のハンドル
	err = dc1394_camera_enumerate (d, &list);		//カメラ一覧を列挙
	if( err != DC1394_SUCCESS )						//エラー処理
	{
	    fprintf( stderr, "Unable to look for cameras\n\n"
		     "Please check \n"
		     "  - if the kernel modules `ieee1394',`raw1394' and `ohci1394' "
		     "are loaded \n"
		     "  - if you have read/write access to /dev/raw1394\n\n");
	    return 1;
	}

    if(list->num == 0)								//カメラの数が0の場合
	{
	    fprintf( stderr, "No cameras found!\n");
	    return 1;
	}

    printf( "There were %d camera(s) found attached to your PC\n", list->num );		//カメラが見つかった場合

   for( nThisCam = 0; nThisCam < list->num; nThisCam++ )							//カメラを探す
	{
	    camera = dc1394_camera_new(d, list->ids[nThisCam].guid);					//使用しているカメラのハンドル

	    if( !camera )
		{
		    printf("Failed to initialize camera with guid %llx and %llx", list->ids[nThisCam].guid, list->ids[nThisCam + 1].guid);
		    continue;
		}

	    printf( "Camera %d model = '%s'\n", nThisCam, camera->model );
	    if( isStereoCamera(camera) )
		{
		    printf( "Using this camera\n" );
		    break;
		}
	    dc1394_camera_free(camera);
	}

    if( nThisCam == list->num )
	{
	    printf( "No stereo cameras were detected\n" );
	    return 0;
	}

    dc1394_camera_free_list (list);						//指定されたGUIDのカメラが見つからなかった場合

    err = queryStereoCamera( camera, &stereoCamera);

    if ( err != DC1394_SUCCESS )
	{
	    fprintf( stderr, "Cannot query all information from camera\n" );
	    cleanup_and_exit( camera );
	}

    if ( stereoCamera.nBytesPerPixel != 2 && stereoCamera2.nBytesPerPixel != 2 )
	{
	    fprintf( stderr, "Example has not been updated to work with XB3 in 3 camera mode yet!\n" );
	    cleanup_and_exit( stereoCamera.camera );
	}

    printf( "Setting stereo video capture mode\n" );    
//printf("%d\n",stereoCamera.nRows);   	
	dc1394_reset_bus(camera);	//リセット
    err = setStereoVideoCapture( &stereoCamera );
    if ( err != DC1394_SUCCESS )
	{
	    fprintf( stderr, "Could not set up video capture mode\n" );
	    //cleanup_and_exit( stereoCamera.camera );
		fprintf( stderr, "Maybe an unclean shutdown was occured in the previous camera capturing.\nTring to reinitialize the bus...\n");
    	//dc1394_reset_bus(camera);	//リセット
    	//sleep(1);					// ちょっと待つ
		cleanup_and_exit( stereoCamera.camera );
	}
    printf( "Start transmission\n" );

    err = startTransmission( &stereoCamera );
    if ( err != DC1394_SUCCESS )
	{
	    fprintf( stderr, "Unable to start camera iso transmission\n" );
	    cleanup_and_exit( stereoCamera.camera );
	}
	printf( "Giving auto-gain algorithm a chance to stabilize\n" );

	nBufferSize = stereoCamera.nRows * stereoCamera.nCols * stereoCamera.nBytesPerPixel;
	pucDeInterlacedBuffer = new unsigned char[ nBufferSize ];
	pucRGBBuffer 	= new unsigned char[ nBufferSize * 3 ];
	pucGreenBuffer 	= new unsigned char[ nBufferSize ];

    sleep(1);
    return 0;

}


int Bumblebee2Control::triclopsInit(){

    printf( "Getting TriclopsContext from camera (slowly)... \n" );
    e = getTriclopsContextFromCamera( &stereoCamera, &triclops );
    if ( e != TriclopsErrorOk )
	{
	    fprintf( stderr, "Can't get context from camera\n" );
	    cleanup_and_exit( camera );
	    return 1;
	}
    printf( "...done\n" );

/*
   if ( !triclopsSetup( triclops, HEIGHT, WIDTH ) )
   {
      printf( "triclopsSetup() failed\n" );
      exit( 1 );
   }
*/
    // make sure we are in subpixel mode
	triclopsSetResolution( triclops , STEREO_HEIGHT, STEREO_WIDTH );

    // make sure we are in subpixel mode
    triclopsSetSubpixelInterpolation( triclops, 1 );
	triclopsSetStrictSubpixelValidation( triclops, 1 );

	triclopsSetDisparity(triclops, 1, 100);					//撮影距離レンジ

	triclopsSetTextureValidation(triclops, 1);	
	triclopsSetTextureValidationThreshold(triclops, 2.0);

	triclopsSetStereoMask(triclops, 13);
	triclopsSetEdgeCorrelation(triclops, 1);				//エッジ処理を有効にし物体抽出精度を高める
	triclopsSetEdgeMask(triclops, 11);

	triclopsSetSurfaceValidation(triclops, 1);				//ノイズ減少モードの有効化
	triclopsSetSurfaceValidationSize(triclops, 250);		//100〜500	
	triclopsSetSurfaceValidationDifference(triclops, 1.0); 

	triclopsSetUniquenessValidation(triclops, 0);
	triclopsSetUniquenessValidationThreshold(triclops, 2.0);	

	triclopsSetLowpass(triclops, 1);

	//ステレオ視の精度設定．TriStereoQlty_ENHANCEDは高精度だが，処理遅くなる. デフォルトはTriStereoQlty_STANDARD
	//triclopsSetStereoQuality(triclops, TriStereoQlty_ENHANCED);
	triclopsSetStereoQuality(triclops, TriStereoQlty_STANDARD);

	float focallength, centerRow, centerCol, baseline;
	TriclopsStereoQuality quality;	//ステレオ視の精度

	triclopsGetFocalLength(triclops, &focallength);
	triclopsGetImageCenter(triclops, &centerRow, &centerCol);
	triclopsGetBaseline(triclops, &baseline);
	triclopsGetStereoQuality(triclops, &quality);

	cout<<"\n---Stereo Setting--- \n"
		"Focal Length "<<focallength<<" pixels for "<<WIDTH<<" x "<<HEIGHT<<"\n"
		"Image Center "<<centerRow<<" "<<centerCol<<"\n"
		"Baseline "<<baseline<<endl;
	if(quality == 0) cout<<"Stereo Quality : Standard"<<endl;
	if(quality == 1) cout<<"Stereo Quality : Enhanced"<<endl;
	cout<<"--------------------\n"<<endl;

    return 0;
}


int Bumblebee2Control::triclopsProcess()
{
	e = triclopsRectify( triclops, &input );
	if ( e != TriclopsErrorOk )
	{
		fprintf( stderr, "triclopsRectify failed!\n" );
		triclopsDestroyContext( triclops );
		cleanup_and_exit( camera );
		return 1;
	}

	e = triclopsStereo( triclops );
	if ( e != TriclopsErrorOk )
	{
		fprintf( stderr, "triclopsStereo failed!\n" );
		triclopsDestroyContext( triclops );
		cleanup_and_exit( camera );
		return 1;
	}
	return 0;
}


int Bumblebee2Control::getColorImage(IplImage *left, IplImage *right){

	unsigned char* pucRightRGB = NULL;
	unsigned char* pucLeftRGB	= NULL;
	unsigned char* pucCenterRGB = NULL;


	extractImagesColor( &stereoCamera,DC1394_BAYER_METHOD_NEAREST, pucDeInterlacedBuffer, pucRGBBuffer, pucGreenBuffer, &pucRightRGB, &pucLeftRGB, &pucCenterRGB, &input );

	int bpp = ((right->depth & 255) / 8) * right->nChannels;
	memcpy(left->imageData, pucLeftRGB, stereoCamera.nRows * stereoCamera.nCols * bpp);	
	memcpy(right->imageData, pucRightRGB, stereoCamera.nRows * stereoCamera.nCols * bpp);	
	return 0;
}


int Bumblebee2Control::getMonoImage(IplImage *left, IplImage *right){

    unsigned char* pucRightMono = NULL;
    unsigned char* pucLeftMono = NULL;
    unsigned char* pucCenterMono = NULL;
    
	extractImagesMono( &stereoCamera,  pucDeInterlacedBuffer,  &pucRightMono,
					   &pucLeftMono,  &pucCenterMono,  &input );
	memcpy(left->imageData, pucLeftMono, stereoCamera.nRows * stereoCamera.nCols );	
	memcpy(right->imageData, pucRightMono, stereoCamera.nRows * stereoCamera.nCols); 	   
						 
	return 0;
}


int Bumblebee2Control::triclopsGetDisp(short int *disp){
//		triclopsGetImage( triclops, TriImg_RECTIFIED, TriCam_REFERENCE, &image );
		triclopsGetImage16( triclops, TriImg16_DISPARITY, TriCam_REFERENCE, &image16 );

		for(int i = 0; i < image16.nrows * image16.ncols; i++){
			disp[i] = image16.data[i];
		}
		return 0;
}


int Bumblebee2Control::triclopsGetPoint(double *x_point, double *y_point, double *z_point)
{
	int x, y;
	int i = 0;
	float px, py, pz;
	unsigned short *depth;
	
	for(y = 0; y < image16.nrows; y++){
		depth = image16.data + y * image16.rowinc / sizeof(unsigned short);
	
	  	for(x = 0; x < image16.ncols; x++, depth++){
			if((0xFF00 & *depth) != 0xFF00){
				triclopsRCD16ToWorldXYZ(triclops, y, x, *depth, &px, &py, &pz);
	    	}
			else{
				px = 0.0, py = 0.0, pz = 0.0;
	    	}
				x_point[i] = (double)px;
				y_point[i] = (double)py;
				z_point[i] = (double)pz;
				i++;
		  //printf("%f, %f, %f\n", px, py, pz);
		}
	}

	return 0;
}


int Bumblebee2Control::RGBtoGL(IplImage *image, double *xxx, double *yyy, double *zzz, double xc, double yc, double zc)
{
    IplImage *imgResize	= cvCreateImage(cvSize(STEREO_WIDTH, STEREO_HEIGHT), IPL_DEPTH_8U, 3);
	cvResize(image, imgResize, CV_INTER_NN);
	int i = 0;

	x_center = xc;
	y_center = yc;
	z_center = zc;

	for(int y = 0; y < STEREO_HEIGHT; y++){
		for(int x = 0; x < STEREO_WIDTH; x++){

		xx_g[i] = xxx[i];
		yy_g[i] = -yyy[i];
		zz_g[i] = zzz[i];

		RR[i] = (int)imgResize->imageData[imgResize->widthStep * y + x * 3 + 2];
 		GG[i] = (int)imgResize->imageData[imgResize->widthStep * y + x * 3 + 1];
 		BB[i] = (int)imgResize->imageData[imgResize->widthStep * y + x * 3 + 0];

		i++;
		}
	}
	return 0;
}


IplImage *Bumblebee2Control::Calibration(IplImage *image, IplImage *imageRect)
{
	FILE *fp;
	float tp[3], cameraMatrix[9], distortion[4];
	CvMat *intrinsic_matrix = cvCreateMat(3, 3, CV_64FC1);		//内部行列
	CvMat *distortion_coeffs = cvCreateMat(4, 1, CV_64FC1);		//歪み

		//内部パラメータのデータがある場合利用する
	if((fp = fopen("calib.txt", "r")) == NULL){
		//printf("File cannot open\n");
		cvCopy(image, imageRect, 0);
		return imageRect;
	}else{
		//printf("File open success!\n");
		
		//printf("内部パラメータ\n");
		
		for(int i = 0; i < 3; i++){
			fscanf(fp,"%f %f %f \n", &tp[0], &tp[1], &tp[2]);
			cameraMatrix[3 * i] = tp[0];
			cameraMatrix[3 * i + 1] = tp[1];
			cameraMatrix[3 * i + 2] = tp[2];
       
			//printf("%7.7lf %7.7lf %7.7lf \n", tp[0], tp[1], tp[2]);
		}
		fscanf(fp,"%f %f %f %f", &distortion[0], &distortion[1], &distortion[2], &distortion[3]);
		//printf("\n");
		//printf("歪み\n");
		//printf("%7.7lf %7.7lf %7.7lf %7.7lf\n", distortion[0], distortion[1], distortion[2], distortion[3]);
		//printf("\n");

		//読み込んだ値を格納
		cvmSet(intrinsic_matrix, 0, 0, cameraMatrix[0]); 
		cvmSet(intrinsic_matrix, 0, 1, cameraMatrix[1]);
		cvmSet(intrinsic_matrix, 0, 2, cameraMatrix[2]);
		cvmSet(intrinsic_matrix, 1, 0, cameraMatrix[3]);
		cvmSet(intrinsic_matrix, 1, 1, cameraMatrix[4]);
		cvmSet(intrinsic_matrix, 1, 2, cameraMatrix[5]);
		cvmSet(intrinsic_matrix, 2, 0, cameraMatrix[6]);
		cvmSet(intrinsic_matrix, 2, 1, cameraMatrix[7]);
		cvmSet(intrinsic_matrix, 2, 2, cameraMatrix[8]);

		cvmSet(distortion_coeffs, 0, 0, distortion[0]);
		cvmSet(distortion_coeffs, 1, 0, distortion[1]);
		cvmSet(distortion_coeffs, 2, 0, distortion[2]);
		cvmSet(distortion_coeffs, 3, 0, distortion[3]);

		fclose(fp);
	}

	cvUndistort2(image, imageRect, intrinsic_matrix, distortion_coeffs);	

	return imageRect;
}


int Bumblebee2Control::Distance(int levels, double *z_point, double *z_point_back, IplImage *mask, short key)
{
	double diff;
	//for(int i = 0 ; i < WIDTH * HEIGHT; i++){
	//	printf("%f, %f\n", z_point[i], z_point_back[i]);
		//}

	if(key == ' '){
		for(int i = 0; i < WIDTH * HEIGHT; i++){
			z_point_back[i] = z_point[i];
		//cout<<" "<<z_point_back[i]<<endl;
		}
	}
	//printf("%d\n", frame);
	
	//背景とカメラ画像の「距離」を計算		
	for(int i = 0; i < WIDTH * HEIGHT; i++){
		//cout<<z_point[i]<<" "<<z_point_back[i]<<endl;		
		//z_point_back[i] = (255.0 - back->imageData[i]) * 8000.0 / 1000.0 / 255.0;
		diff = fabs(z_point[i] - z_point_back[i]);
   
		if(z_point[i] > 100.0)
			mask->imageData[i] = 0;

		if(z_point_back[i] > 100.0)
			mask->imageData[i] = 0;

		if(diff > z_point[i] || diff > z_point_back[i])
			mask->imageData[i] = 0;			
			
		//else if(diff > z_point_back[i] * 0.8)
			//mask->imageData[i] = 0;

		//else if(diff < z_point[i] * levels / 100)
			//mask->imageData[i] = 0;  //背景

		if(diff < z_point_back[i] * levels / 100)
			mask->imageData[i] = 0;  //背景

		else if(diff == 0)
			mask->imageData[i] = 0;  //背景
		//else if(diff < 0.50)
			//mask->imageData[i] = 0;  //背景

		//else if((z_point_back[i] > 8.0) || (z_point_back[i] < 0.2) && z_point_back[i] != 0.0)			
			//mask->imageData[i] = 0;  //背景

		//if((z_point[i] > 6.0) || (z_point[i] < 0.2))
			//mask->imageData[i] = 0;  //背景
		else
			mask->imageData[i] = 1;  //前景

		}
		return 0;
}

/*
int Bumblebee2Control::InitBrightness(IplImage *right, IplImage *av_img, IplImage *sgm_img, IplImage *tmp_img)
{
	int INIT_TIME = 100;
	int i;//, counter;

	cvSetZero(av_img);
	while(i == INIT_TIME){			
		//cvUndistort2(right, rightRect, intrinsic_matrix, distortion_coeffs);
	
		cvAcc(right, av_img);
		i++;
	}
	cvConvertScale(av_img, av_img, 1.0 / INIT_TIME);

	cvSetZero(sgm_img);
	while(i == INIT_TIME){
		//cvUndistort2(right, rightRect, intrinsic_matrix, distortion_coeffs);

    	cvConvert(right, tmp_img);
    	cvSub(tmp_img, av_img, tmp_img);
    	cvPow(tmp_img, tmp_img, 2.0);
    	cvConvertScale(tmp_img, tmp_img, 2.0);
    	cvPow(tmp_img, tmp_img, 0.5);
    	cvAcc(tmp_img, sgm_img);
		i++;
	}
	cvConvertScale(sgm_img, sgm_img, 1.0 / INIT_TIME);

	//counter = 0;
	return 0;
	
}
*/

int Bumblebee2Control::InitBrightness2(IplImage *right)
{
	if(right){
		//av_img = NULL;
		//sgm_img = NULL;
		//tmp_img = NULL;
		av_img = cvCreateImage(cvSize(WIDTH, HEIGHT), IPL_DEPTH_32F, 3);
		sgm_img = cvCreateImage(cvSize(WIDTH, HEIGHT), IPL_DEPTH_32F, 3);
		tmp_img = cvCreateImage(cvSize(WIDTH, HEIGHT), IPL_DEPTH_32F, 3);
		lower_img = cvCreateImage(cvSize(WIDTH, HEIGHT), IPL_DEPTH_32F, 3);
		upper_img = cvCreateImage(cvSize(WIDTH, HEIGHT), IPL_DEPTH_32F, 3);
	}

	int INIT_TIME = 100;
	int i;//, counter;

	cvSetZero(av_img);
	while(i == INIT_TIME){			
		//cvUndistort2(right, rightRect, intrinsic_matrix, distortion_coeffs);
	
		cvAcc(right, av_img);
		i++;
	}
	cvConvertScale(av_img, av_img, 1.0 / INIT_TIME);

	cvSetZero(sgm_img);
	while(i == INIT_TIME){
		//cvUndistort2(right, rightRect, intrinsic_matrix, distortion_coeffs);

    	cvConvert(right, tmp_img);
    	cvSub(tmp_img, av_img, tmp_img);
    	cvPow(tmp_img, tmp_img, 2.0);
    	cvConvertScale(tmp_img, tmp_img, 2.0);
    	cvPow(tmp_img, tmp_img, 0.5);
    	cvAcc(tmp_img, sgm_img);
		i++;
	}
	cvConvertScale(sgm_img, sgm_img, 1.0 / INIT_TIME);
	return 0;
	
}


int Bumblebee2Control::Brightness2(IplImage *right, IplImage *dst_img, IplImage *mask_b)
{
	double B_PARAM = 1.0 / 50.0;//50.0;
	double T_PARAM = 1.0 / 220.0;//200.0;
	double Zeta = 10.0;

	cvConvert(right, tmp_img);

	cvSub(av_img, sgm_img, lower_img);
    cvSubS(lower_img, cvScalarAll(Zeta), lower_img);
    cvAdd(av_img, sgm_img, upper_img);
    cvAddS(upper_img, cvScalarAll(Zeta), upper_img);
    cvInRange(tmp_img, lower_img, upper_img, mask_b);

    //輝度振幅を再計算する
    cvSub(tmp_img, av_img, tmp_img);
    cvPow(tmp_img, tmp_img, 2.0);
    cvConvertScale(tmp_img, tmp_img, 2.0);
    cvPow(tmp_img, tmp_img, 0.5);

    //背景と判断された領域の背景の輝度平均と輝度振幅を更新する
    cvRunningAvg(right, av_img, B_PARAM, mask_b);
    cvRunningAvg(tmp_img, sgm_img, B_PARAM, mask_b);

    //物体領域と判断された領域では輝度振幅のみを（背景領域よりも遅い速度で）更新する
    cvNot(mask_b, mask_b);
    cvRunningAvg(tmp_img, sgm_img, T_PARAM, mask_b);

    //物体領域のみを出力画像にコピーする（背景領域は黒）
    cvSetZero(dst_img);
	return 0;
}

/*
int Bumblebee2Control::Brightness(IplImage *right, IplImage *av_img, IplImage *sgm_img, IplImage *tmp_img, IplImage *lower_img, IplImage *upper_img, IplImage *dst_img, IplImage *mask2)
{
	double B_PARAM = 1.0 / 50.0;//50.0;
	double T_PARAM = 1.0 / 220.0;//200.0;
	double Zeta = 10.0;

	cvConvert(right, tmp_img);

	cvSub(av_img, sgm_img, lower_img);
    cvSubS(lower_img, cvScalarAll(Zeta), lower_img);
    cvAdd(av_img, sgm_img, upper_img);
    cvAddS(upper_img, cvScalarAll(Zeta), upper_img);
    cvInRange(tmp_img, lower_img, upper_img, mask2);

    //輝度振幅を再計算する
    cvSub(tmp_img, av_img, tmp_img);
    cvPow(tmp_img, tmp_img, 2.0);
    cvConvertScale(tmp_img, tmp_img, 2.0);
    cvPow(tmp_img, tmp_img, 0.5);

    //背景と判断された領域の背景の輝度平均と輝度振幅を更新する
    cvRunningAvg(right, av_img, B_PARAM, mask2);
    cvRunningAvg(tmp_img, sgm_img, B_PARAM, mask2);

    //物体領域と判断された領域では輝度振幅のみを（背景領域よりも遅い速度で）更新する
    cvNot(mask2, mask2);
    cvRunningAvg(tmp_img, sgm_img, T_PARAM, mask2);

    //物体領域のみを出力画像にコピーする（背景領域は黒）
    cvSetZero(dst_img);

    //counter++;
	return 0;
}
*/

int Bumblebee2Control::Coordinate(int x, int y, float &px, float &py, float &pz)
{
	float X, Y, Z;
	float X_, Y_, Z_;		//スピーカ座標系
	float Xs, Ys, Zs;		//四捨五入
	int pixelinc;
	unsigned short *row;
	unsigned short disparity;
	//unsigned short int *depth;
	//char buf[2048];		//add
	CvPoint pt;
	
	//triclopsGetImage16( triclops, TriImg16_DISPARITY, TriCam_REFERENCE, &image16 );

	pixelinc = image16.rowinc / 2;

	pt = cvPoint(x, y);

	row = image16.data + y * pixelinc;
	disparity = row[x];

	if(disparity < 0xFF00){	
		triclopsRCD16ToXYZ(triclops, y, x, disparity, &X, &Y, &Z);

		if(Z < 5.0 && Z > 0.0){
			//printf("画像座標(pixel): %d %X_d\n", x, y);    //画像座標取得
			//printf("ワールド座標(m): %f %f %f\n",X, -Y, Z);
		
			change(X, Y, Z, X_, Y_, Z_);			//スピーカアレイ系に座標変換
			//X_ = 4.0 - 1.0 / sqrt(2.0) * (Z - Y);
			//Y_ = X + 2.0;
			//Z_ = h - 1.0 / sqrt(2.0) * (Y + Z);

			Xs = HalfAdjust(X_) * 1000.0;//X_*1000.0;			//四捨五入
			Ys = HalfAdjust(Y_) * 1000.0;//Y_*1000.0;//
			Zs = HalfAdjust(Z_) * 1000.0;//Z_*1000.0;//
			//printf("スピーカアレイ座標(m): %f %f %f\n", Xs, Ys, Zs);

				//Xs = X_ * 1000.0;
				//Ys = Y_ * 1000.0;
				//Zs = Z_ * 1000.0;
		}
	}

	else if(disparity >= 0xFF00){
		//printf("--------------------------\n");
		//printf("Nothing!\n");
		//printf("--------------------------\n");
		//printf("画像座標(pixel): %d %d\n", x, y);    //画像座標取得	

	  	bool b = true;
	  	int n = 1;
	  	while(b || n == 5){

			for(int itmp = 0; itmp < 8; itmp++){
				if(DoesExistZ(disparity, row, x, y, X, Y, Z, n, itmp, Xs, Ys, Zs)){
					b = false;
					break;
		  		}
	  		}
	  		n++;
		}
	}

	else printf("?\n");
	
	//printf("%d\n", p);
	//cvShowImage(window_rectify, circle);

	px = Xs;//2000.0;
	py = Ys;
	pz = Zs;

	return 0;
}


int Bumblebee2Control::Coordinate2(int x, int y, float &px, float &py, float &pz)
{
	float X, Y, Z;
	float X_, Y_, Z_;		//スピーカ座標系
	float Xs, Ys, Zs;		//四捨五入
	int pixelinc;
	unsigned short *row;
	unsigned short disparity;
	//unsigned short int *depth;
	//char buf[2048];		//add
	CvPoint pt;
	
	//triclopsGetImage16( triclops, TriImg16_DISPARITY, TriCam_REFERENCE, &image16 );

	pixelinc = image16.rowinc / 2;

	pt = cvPoint(x, y);

	row = image16.data + y * pixelinc;
	disparity = row[x];

	if(disparity < 0xFF00){	
		triclopsRCD16ToXYZ(triclops, y, x, disparity, &X, &Y, &Z);

		if(Z < 5.0 && Z > 0.0){
			//printf("画像座標(pixel): %d %X_d\n", x, y);    //画像座標取得
			//printf("ワールド座標(m): %f %f %f\n",X, -Y, Z);
		
			//change(X, Y, Z, X_, Y_, Z_);			//スピーカアレイ系に座標変換
			//X_ = 4.0 - 1.0 / sqrt(2.0) * (Z - Y);
			//Y_ = X + 2.0;
			//Z_ = h - 1.0 / sqrt(2.0) * (Y + Z);

			Xs = HalfAdjust(X) * 1000.0;//X_*1000.0;			//四捨五入
			Ys = HalfAdjust(Y) * 1000.0;//Y_*1000.0;//
			Zs = HalfAdjust(Z) * 1000.0;//Z_*1000.0;//
			//printf("スピーカアレイ座標(m): %f %f %f\n", Xs, Ys, Zs);

				//Xs = X_ * 1000.0;
				//Ys = Y_ * 1000.0;
				//Zs = Z_ * 1000.0;
		}
	}

	else if(disparity >= 0xFF00){
		//printf("--------------------------\n");
		//printf("Nothing!\n");
		//printf("--------------------------\n");
		//printf("画像座標(pixel): %d %d\n", x, y);    //画像座標取得	

	  	bool b = true;
	  	int n = 1;
	  	while(b || n == 5){

			for(int itmp = 0; itmp < 8; itmp++){
				if(DoesExistZ2(disparity, row, x, y, X, Y, Z, n, itmp, Xs, Ys, Zs)){
					b = false;
					break;
		  		}
	  		}
	  		n++;
		}
	}

	else printf("?\n");
	
	//printf("%d\n", p);
	//cvShowImage(window_rectify, circle);

	px = Xs;//2000.0;
	py = Ys;
	pz = Zs;

	return 0;
}


double Bumblebee2Control::timer_ms()
{
	struct timeval tv;
	gettimeofday(&tv, NULL);
	return ((tv.tv_sec + tv.tv_usec * 1e-6) * 1e3);
}


float Bumblebee2Control::HalfAdjust(float c)
{
	int tmp = (int)((c + 0.0005) * 1000);
	return (float)tmp / 1000.0;
}


float Bumblebee2Control::change(float x, float y, float z, float &pxx, float &pyy, float &pzz)
{
	float theta, rad;		//カメラ角度
	float h;				//カメラ設置位置の高さ
	float px, py, pz;
	CvMat *changeMatrix = cvCreateMat(3, 3, CV_64FC1);		//3行3列
	CvMat *dMatrix = cvCreateMat(3, 1, CV_64FC1);			//3行1列
	CvMat *Matrix = cvCreateMat(3, 1, CV_64FC1);			//3行1列

	//theta = -50.0;//-46.0;		//degree
	theta = 0 - _theta;
	rad = theta * (double)M_PI / 180.0;
	h = 2.730;			//m

	cvmSet(changeMatrix, 0, 0, 1);
	cvmSet(changeMatrix, 0, 1, 0);
	cvmSet(changeMatrix, 0, 2, 0);
	cvmSet(changeMatrix, 1, 0, 0);
	cvmSet(changeMatrix, 1, 1, cos(rad));
	cvmSet(changeMatrix, 1, 2, -sin(rad));
	cvmSet(changeMatrix, 2, 0, 0);
	cvmSet(changeMatrix, 2, 1, sin(rad));
	cvmSet(changeMatrix, 2, 2, cos(rad));

	cvmSet(dMatrix, 0, 0, x);
	cvmSet(dMatrix, 1, 0, y);
	cvmSet(dMatrix, 2, 0, z);

	cvmMul(changeMatrix, dMatrix, Matrix);

	px = (float)cvmGet(Matrix, 0, 0);
	py = (float)cvmGet(Matrix, 1, 0);
	pz = (float)cvmGet(Matrix, 2, 0);

	pxx = 4.0 - pz;
	pyy = px + 2.0;
	pzz = h - py;

	return 0;
}


bool Bumblebee2Control::DoesExistZ(unsigned short disparity, unsigned short *row, int x , int y , float X , float Y , float Z , int n, int type, float &px, float &py, float &pz)
{
	int x_ = 0 , y_ = 0;
	char ctmp_x = '+' ,ctmp_y = '+';
	//char buf[2048];
	//unsigned short *row;
	int pixelinc;
	float X_, Y_, Z_;		//スピーカ座標系
	float Xs, Ys, Zs;		//四捨五入
	//row = image16.data + y * pixelinc;
	//triclopsGetImage16(triclops, TriImg16_DISPARITY, TriCam_REFERENCE, &image16);
 	pixelinc = image16.rowinc / 2;

	switch( type ){
		case 0:
		case 1:
	  			x_ = n;
	  			if( type == 1 )	y_ = -n;
	  			break;
		case 2:
		case 3:
				y_ = -n;
	  			if( type == 3 ) x_ = -n;
	  			break;
		case 4:
		case 5:
			  	x_ = -n;
	  			if( type == 5 ) y_ = n;
	  			break;
		case 6:
		case 7:
				y_ = n;
	  			if( type == 7 ) x_ = n;
	  			break;
	}
	if( x_ < 0 ) ctmp_x = '-';
	if( y_ < 0 ) ctmp_y = '-';

	if(disparity >= 0xFF00){
		row = image16.data + (y + y_) * pixelinc; 
		//printf("%d;%d + %d = %d\n",n,x,x_,x+x_);
		disparity = row[x + x_];

		if(disparity < 0xFF00){	
	  		triclopsRCD16ToXYZ(triclops, y+(y_), x+(x_), disparity, &X, &Y, &Z);

		  	if(Z < 5.0 && Z > 0.0){
				//printf("周辺画像座標(pixel): %d %d\n", x + x_, y + y_);    //画像座標取得
				//printf("ワールド座標(m): %f %f %f (x%c%d, y%c%d)\n",X, -Y, Z, ctmp_x , x_ , ctmp_y , y_);

				change(X, Y, Z, X_, Y_, Z_);			//スピーカアレイ系に座標変換
				//X_ = 4.0 - 1.0 / sqrt(2.0) * (Z - Y);
				//Y_ = X + 2.0;
				//Z_ = h - 1.0 / sqrt(2.0) * (Y + Z);
				
				Xs = HalfAdjust(X_) * 1000.0;	//X_*1000.0;		//四捨五入
				Ys = HalfAdjust(Y_) * 1000.0; //Y_*1000.0;
				Zs = HalfAdjust(Z_) * 1000.0; //Z_*1000.0;
				//Xs = X_ * 1000.0;
				//Ys = Y_ * 1000.0;
				//Zs = Z_ * 1000.0;
			
				px = Xs;//2000.0;
				py = Ys;	
				pz = Zs;			

				return true;
			}

	  		else if(disparity >= 0xFF00){
	  			printf("**************************\n");
	  			printf("Please give it up!\n");
	  			printf("**************************\n");
	  			return false;
			}		
		}
	}

  	return false;
}


bool Bumblebee2Control::DoesExistZ2(unsigned short disparity, unsigned short *row, int x , int y , float X , float Y , float Z , int n, int type, float &px, float &py, float &pz)
{
	int x_ = 0 , y_ = 0;
	char ctmp_x = '+' ,ctmp_y = '+';
	//char buf[2048];
	//unsigned short *row;
	int pixelinc;
	float X_, Y_, Z_;		//スピーカ座標系
	float Xs, Ys, Zs;		//四捨五入
	//row = image16.data + y * pixelinc;
	//triclopsGetImage16(triclops, TriImg16_DISPARITY, TriCam_REFERENCE, &image16);
 	pixelinc = image16.rowinc / 2;

	switch( type ){
		case 0:
		case 1:
	  			x_ = n;
	  			if( type == 1 )	y_ = -n;
	  			break;
		case 2:
		case 3:
				y_ = -n;
	  			if( type == 3 ) x_ = -n;
	  			break;
		case 4:
		case 5:
			  	x_ = -n;
	  			if( type == 5 ) y_ = n;
	  			break;
		case 6:
		case 7:
				y_ = n;
	  			if( type == 7 ) x_ = n;
	  			break;
	}
	if( x_ < 0 ) ctmp_x = '-';
	if( y_ < 0 ) ctmp_y = '-';

	if(disparity >= 0xFF00){
		row = image16.data + (y + y_) * pixelinc; 
		//printf("%d;%d + %d = %d\n",n,x,x_,x+x_);
		disparity = row[x + x_];

		if(disparity < 0xFF00){	
	  		triclopsRCD16ToXYZ(triclops, y+(y_), x+(x_), disparity, &X, &Y, &Z);

		  	if(Z < 5.0 && Z > 0.0){
				//printf("周辺画像座標(pixel): %d %d\n", x + x_, y + y_);    //画像座標取得
				//printf("ワールド座標(m): %f %f %f (x%c%d, y%c%d)\n",X, -Y, Z, ctmp_x , x_ , ctmp_y , y_);

				//change(X, Y, Z, X_, Y_, Z_);			//スピーカアレイ系に座標変換
				//X_ = 4.0 - 1.0 / sqrt(2.0) * (Z - Y);
				//Y_ = X + 2.0;
				//Z_ = h - 1.0 / sqrt(2.0) * (Y + Z);
				
				Xs = HalfAdjust(X) * 1000.0;	//X_*1000.0;		//四捨五入
				Ys = HalfAdjust(Y) * 1000.0; //Y_*1000.0;
				Zs = HalfAdjust(Z) * 1000.0; //Z_*1000.0;
				//Xs = X_ * 1000.0;
				//Ys = Y_ * 1000.0;
				//Zs = Z_ * 1000.0;
			
				px = Xs;//2000.0;
				py = Ys;	
				pz = Zs;			

				return true;
			}

	  		else if(disparity >= 0xFF00){
	  			printf("**************************\n");
	  			printf("Please give it up!\n");
	  			printf("**************************\n");
	  			return false;
			}		
		}
	}

  	return false;
}


int Bumblebee2Control::acs(char *buf, int argc, char *p)
{
	int sockfd;
	struct sockaddr_in client_addr;
	int buf_len = 1024;

	//ソケットを生成-----------------------------------------------------------------
	//	int socket(int pf, type, int protocol)										
	//	pf：プロトコルファミリと呼ばれ、プロトコルの種類を特定する
	//	type：コネクション型であるTCP用のソケットを作るためにはSOCK_STREAM
	//		  データダイアグラム型であるUDPを使う場合はSOCK_DGRAMとする
	//	protocol：プロトコルを指定するが、インターネットを前提にする限りは0とする
	//-------------------------------------------------------------------------------
	if((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0){
		perror("client: socket");
		exit(1);
	}
	fprintf(stderr,"4 ");

	//client_addr構造体に、接続するサーバのアドレス・ポート番号を設定----------------
	//	void bzero(void *dst, size_t nbytes) バイト単位の0初期化関数 ゼロで埋める
	//	dst：初期化先(コピー先)の先頭アドレス
	//	nbytes：初期化(コピー)する大きさ(バイト)
	//-------------------------------------------------------------------------------
	bzero((char *)&client_addr, sizeof(client_addr));
	client_addr.sin_family = AF_INET;
	client_addr.sin_addr.s_addr = inet_addr(p);
	client_addr.sin_port = htons(8000);
	fprintf(stderr,"3 ");
    
	//ソケットをサーバに接続---------------------------------------------------------
	//	int connect(int socket, struct sockaddr *destaddr, int destlen)
	//	socket：socket()システムコールが返したソケット番号をいれる
	//	destaddr：接続先のサーバを指定するsockaddr構造体へのポインタ。ただし、
	//			  インターネット通信ではsockaddr構造体ではなく、sockaddr_in構造体を
	//			  sockaddr構造体にキャスト与える
	//-------------------------------------------------------------------------------
	if(connect(sockfd, (struct sockaddr *)&client_addr, sizeof(client_addr)) < 0){
		perror("client: connect");
		close(sockfd);
		return -1;
		//exit(1);
	}

	//ソケットへの送信---------------------------------------------------------------
	//	int write(int socket, char *buf, int length)
	//	socket：ソケット番号
	//	buf：送信すべきデータを格納してある領域
	//	length：送信データのバイト数
	//-------------------------------------------------------------------------------
	write(sockfd, buf, buf_len);
    
	fprintf(stderr,"2 ");

	close(sockfd);
	fprintf(stderr,"1\n");
	
	return 0; 
}


void Bumblebee2Control::RectDetec(IplImage *img, int *_proxmax, int *_proymax, CvPoint *_xy, CvPoint *_xy1)
{
	//#define WIDTH 320
	//#define HEIGHT 240
	int k = 0;
	int j = 0;
	int i;
	CvPoint xy, xy1;
	xy.x = 0;
	xy.y = 0;
	xy1.x = 0;
	xy1.y = 0;
	unsigned char *data = (unsigned char*)img->imageData;
	int proxmax = 0;
	int proymax = 0;
	int promax_x = 0;
	int promax_y = 0;
	int prox[WIDTH] = {};
	int proy[HEIGHT] = {};

	for(int i = 0; i < WIDTH; i++)prox[i] = 0;
	for(int i = 0; i < HEIGHT; i++)proy[i] = 0;

	for(int i = 0; i < WIDTH * HEIGHT; i++){
		if(i > 14){
			if((data[i * 3] != 0) && (data[i * 3 + 1] != 0) && (data[i * 3 + 2] != 0)){
			//	if(histgram[hstable[data[i*3+2]][data[i*3+1]][data[i*3]].S][hstable[data[i*3+2]][data[i*3+1]][data[i*3]].H]>1){
				prox[j] = prox[j] + 1;
				proy[k] = proy[k] + 1;
		//	}
			}
		}
		j++;
		if(j == WIDTH){
			j = 0;
			k++;
		}
	}
	prox[0] = prox[WIDTH-1] = proy[0] = proy[HEIGHT-1] = proxmax = proymax = promax_x = promax_y = 0;

	for(i = 0; i < WIDTH; i++){
		if(prox[i] < 5)  prox[i] = 0;
		if(prox[i] > proxmax) { proxmax = prox[i]; promax_x = i; }
	}

	for(i = 0; i < HEIGHT; i++){
		if(proy[i] < 5)  proy[i] = 0;
		if(proy[i] > proymax) { proymax = proy[i]; promax_y = i; }
	}
	for(i = promax_x; i > 0; i--)   { if(prox[i] <= 10) { xy.x = i; break; } } 
	for(i = promax_x; i < WIDTH; i++) { if(prox[i] <= 10) { xy1.x = i; break; } }
	for(i = promax_y; i > 0; i--)   { if(proy[i] <= 10) { xy.y = i; break; } } 
	for(i = promax_y; i < HEIGHT; i++) { if(proy[i] <= 10) { xy1.y = i; break; } }

	_xy->x = xy.x;
	_xy->y = xy.y;
	_xy1->x =xy1.x;
	_xy1->y = xy1.y;
	*_proxmax = proxmax;
	*_proymax = proymax;
}


IplImage *Bumblebee2Control::IRgBy(IplImage *Synthesis, IplImage *Irgby, IplImage *mask, CvPoint range0, CvPoint range1)
{
	double r, g, b;
	int SCALE = 0;
	int i = 0, j = 0;
	uchar p[3];

	//IplImage *mask = NULL;
	//IplImage *hue_img = NULL, *satu_img = NULL;
	//IplImage *irgby_gray = NULL;
	//IplImage *face = NULL;

/*	double *I, *Rg, *By;
//	double I_pre, Rg_pre, By_pre;
	double *I_med, *Rg_med, *By_med;
	double *hue;
	double *I_filt, *MAD, *MAD_med, *saturation;
	//double Rg_med = 0.0, By_med = 0.0;
*/
/*
	IplImage *I = NULL, *Rg = NULL, *By = NULL;
//	double I_pre, Rg_pre, By_pre;
	IplImage *I_med = NULL, *Rg_med = NULL, *By_med = NULL;
	IplImage *hue = NULL;
	IplImage *I_filt = NULL, *MAD = NULL, *MAD_med = NULL, *saturation = NULL;
*/

	//printf("%d %d %d %d\n", range0.x, range1.x, range0.y, range1.y);
	int W, H;
	//対象人物領域の縦と横の長さを計算 range0 < range1
	//rang0のxあるいはyが0でない場合は1をたさなければならない

	if(range0.x != 0)W = range1.x - range0.x + 1;
	if(range0.y != 0)H = range1.y - range0.y + 1;
	if(range0.x == 0)W = abs(range1.x - range0.x);
	if(range0.y == 0)H = abs(range1.y - range0.y);

	//W = abs(range1.x - range0.x)+1;
	//H = abs(range1.y - range0.y)+1;
	if(range0.x == 0 && range0.y == 0 && range1.x == 0 && range0.y == 0)
		return NULL;
	if(W == 0)
		return NULL;

	//cout<<"W*H"<<W*H<<":W"<<W<<":H"<<H<<endl;
	//printf("%d %d\n", W, H);
	double I_pre, Rg_pre, By_pre, hue_pre;
	CvMat *I = NULL, *Rg = NULL, *By = NULL, *Rg_med = NULL, *By_med = NULL, *hue = NULL;
	CvMat *I_filt = NULL, *MAD = NULL, *MAD_med = NULL, *saturation = NULL;

	I = cvCreateMat(H, W, CV_8UC1);
	Rg = cvCreateMat(H, W, CV_8UC1);
	By = cvCreateMat(H, W, CV_8UC1);
	//CvMat *I_med = cvCreateMat(Hight, Width, CV_8UC1);
	Rg_med = cvCreateMat(H, W, CV_8UC1);
	By_med = cvCreateMat(H, W, CV_8UC1);
	hue = cvCreateMat(H, W, CV_8UC1);
	I_filt = cvCreateMat(H, W, CV_8UC1);
	MAD = cvCreateMat(H, W, CV_8UC1);
	MAD_med = cvCreateMat(H, W, CV_8UC1);
	saturation = cvCreateMat(H, W, CV_8UC1);

	//median = cvCreateImage(cvSize(WIDTH,HEIGHT), IPL_DEPTH_8U, 3);
	//irgby = cvCreateImage(cvSize(WIDTH, HEIGHT), IPL_DEPTH_8U, 3);
	//mask = cvCreateImage(cvSize(WIDTH, HEIGHT), IPL_DEPTH_8U, 1);
	//hue_img = cvCreateImage(cvSize(WIDTH, HEIGHT), IPL_DEPTH_8U, 1);
	//satu_img = cvCreateImage(cvSize(WIDTH, HEIGHT), IPL_DEPTH_8U, 1);
	//irgby_gray = cvCreateImage(cvGetSize(irgby), IPL_DEPTH_8U, 1);
	//face = cvCreateImage(cvSize(WIDTH, HEIGHT), IPL_DEPTH_8U, 3);


	SCALE = (int)((W + H) / 320);		//メディアンフィルタのためのスケール計算 W != 0
	if(SCALE < 1 && SCALE <= 0)SCALE = 1;
	if(SCALE % 2 == 0)SCALE = SCALE + 1;	//OpenCVのメディアンフィルタ関数は偶数ではいけない

	//cout<<SCALE<<endl;
	//ピクセルデータ（R,G,B）を順次取得し，変換する
	for(int y = range0.y; y <= range1.y; y++){
		i = 0;
		for(int x = range0.x; x <= range1.x; x++){
			p[0] = Synthesis->imageData[Synthesis->widthStep * y + x * 3];        // B
			p[1] = Synthesis->imageData[Synthesis->widthStep * y + x * 3 + 1];    // G
			p[2] = Synthesis->imageData[Synthesis->widthStep * y + x * 3 + 2];    // R

			if(p[0] != 0 && p[1] != 0 && p[2] != 0){
				Logopp(p[0], b);
				Logopp(p[1], g);
				Logopp(p[2], r);

				I_pre   = (r + g + b) / 3.0;
				Rg_pre  = r - g;
				By_pre  = b - (g + r) / 2.0;
				hue_pre = atan2(Rg_pre, By_pre) * (180.0 / M_PI);

				//printf("%f\n",Rg_pre);
				cvSetReal2D(I, j, i, I_pre);
				cvSetReal2D(Rg, j, i, Rg_pre);
				cvSetReal2D(By, j, i, By_pre);
				cvSetReal2D(hue, j, i, hue_pre);
				//printf("%d %d %d %d\n", x, y, i, j);
				i++;
			}
			else i++;
		}
		j++;
	}
	//printf("%d %d\n", i, j);

	//自作メディアンフィルタ関数
	//MedianFilter(Rg, Rg_med, WIDTH, HEIGHT, 2 * 2 * SCALE);
	//MedianFilter(By, By_med, WIDTH, HEIGHT, 2 * 2 * SCALE);
	//メディアンフィルタ処理を行うときその範囲は1辺が3*SCALEの正方形となる.ただし,人物領域のサイズを越えてはいけない
	if(3 * 3 * SCALE * SCALE < W * H && 3 * SCALE < W && 3 * SCALE < H){
		cvSmooth(Rg, Rg_med, CV_MEDIAN, 3 * SCALE, 0, 0, 0);	//メディアンフィルタ
		cvSmooth(By, By_med, CV_MEDIAN, 3 * SCALE, 0, 0, 0);
	}
	else{
		cvSmooth(Rg, Rg_med, CV_MEDIAN, SCALE, 0, 0, 0);	//メディアンフィルタ
		cvSmooth(By, By_med, CV_MEDIAN, SCALE, 0, 0, 0);
	}
	
	//Compute texture amplitude
	//MedianFilter(I, I_filt, WIDTH, HEIGHT, 2 * 4 * SCALE);
	if(7 * 7 * SCALE * SCALE < W * H && 7 * SCALE < W && 7 * SCALE < H){
		cvSmooth(I, I_filt, CV_MEDIAN, 7 * SCALE, 0, 0, 0);	
	}
	else{
		cvSmooth(I, I_filt, CV_MEDIAN, SCALE, 0, 0, 0);
	}


	//cvAbsDiff(I, I_filt, MAD);
	//cvPow(Rg_med, Rg_med, 2);
	//cvPow(By_med, By_med, 2);
	//cvAdd(Rg_med, By_med, saturation);

	//hue,saturationを求める
	//printf("here\n");
	for(int y = 0; y < j; y++){
		for(int x = 0; x < i; x++){
			double Rg_data = cvGetReal2D(Rg_med, y, x);
			double By_data = cvGetReal2D(By_med, y, x);
			//printf("%f\n", Rg_data);
			cvSetReal2D(MAD, y, x, cvGetReal2D(I, y, x) - cvGetReal2D(I_filt, y, x));
			cvSetReal2D(MAD, y, x, fabs(cvGetReal2D(MAD, y, x)));
			cvSetReal2D(saturation, y, x, sqrt(Rg_data * Rg_data + By_data * By_data));
			//cvSetReal2D(saturation, y, x, cvSqrt(cvGetReal2D(saturation, y, x)));
			
			//hue_img->imageData[hue_img->widthStep * y + x] = (unsigned char)cvGetReal2D(hue, y, x);
			//satu_img->imageData[satu_img->widthStep * y + x] = (unsigned char)cvGetReal2D(saturation, y, x);
		}
	}

 	//MAD_med = cvCloneImage(MAD);
	//MedianFilter(MAD, MAD_med, WIDTH, HEIGHT, 2 * 6 * SCALE);
	if(11 * 11 * SCALE * SCALE < W * H && 11 * SCALE < W && 11 * SCALE < H){
		cvSmooth(MAD, MAD_med, CV_MEDIAN, 11 * SCALE, 0, 0, 0);
	}
	else{
		cvSmooth(MAD, MAD_med, CV_MEDIAN, SCALE, 0, 0, 0);
	}

	i = 0;
	j = 0;
	//printf("here2\n");
	//マスク画像の作成
	cvSetZero(mask);	//ピクセル値をすべて0にする
	for(int y = range0.y; y <= range1.y; y++){
		i = 0;
		for(int x = range0.x; x <= range1.x; x++){
			double MAD_data = cvGetReal2D(MAD_med, j, i);
			double hue_data = cvGetReal2D(hue, j, i);
			double saturation_data = cvGetReal2D(saturation, j, i);
			
			if(MAD_data < 4.5 && 120.0 < hue_data && hue_data < 160.0 && 10.0 < saturation_data && saturation_data < 60.0){
				mask->imageData[mask->widthStep * y + x] = (unsigned char)1;//printf("clear\n");
			}

			else if(MAD_data < 4.5 && 150.0 < hue_data && hue_data < 180.0 && 20.0 < saturation_data && saturation_data < 80.0){
				mask->imageData[mask->widthStep * y + x] = (unsigned char)1;//printf("clear\n");
			
			}
			i++;
		}
		j++;
	}

	cvDilate(mask, mask, NULL, 5);	//膨張処理
	//printf("here3\n");
	i = 0;
	j = 0;
	//マスク画像の作成
	for(int y = range0.y; y <= range1.y; y++){
		i = 0;
		for(int x = range0.x; x <= range1.x; x++){
			double hue_data = cvGetReal2D(hue, j, i);
			double saturation_data = cvGetReal2D(saturation, j, i);
			if((int)(mask->imageData[mask->widthStep * y + x]) == 1 && 110.0 <= hue_data && hue_data <= 180.0 && 0.0 <= saturation_data && saturation_data <= 130.0){
				mask->imageData[mask->widthStep * y + x] = (unsigned char)1;//printf("clear\n");
			}
			else mask->imageData[mask->widthStep * y + x] = (unsigned char)0;
			i++;
		}
		j++;
	}

	memset(Irgby->imageData, 0, WIDTH * HEIGHT * 3);
	cvCopy(Synthesis, Irgby, mask);
	//FaceDetect(irgby, mask, face);
/*
	cvReleaseMat(&I);
	cvReleaseMat(&Rg);
	cvReleaseMat(&By);
	//cvReleaseMat(&I_med);
	cvReleaseMat(&Rg_med);
	cvReleaseMat(&By_med);
	cvReleaseMat(&I_filt);
	cvReleaseMat(&hue);
	cvReleaseMat(&MAD);
	cvReleaseMat(&MAD_med);
	cvReleaseMat(&saturation);
*/
	//cvReleaseImage(&mask);
	//cvReleaseImage(&hue_img);
	//cvReleaseImage(&satu_img);
/*
	cvInitMatHeader(I, H, W, CV_8UC1);
	cvInitMatHeader(Rg, H, W, CV_8UC1);
	cvInitMatHeader(By, H, W, CV_8UC1);
	cvInitMatHeader(Rg_med, H, W, CV_8UC1);
	cvInitMatHeader(By_med, H, W, CV_8UC1);
	cvInitMatHeader(I_filt, H, W, CV_8UC1);
	cvInitMatHeader(hue, H, W, CV_8UC1);
	cvInitMatHeader(MAD, H, W, CV_8UC1);
	cvInitMatHeader(MAD_med, H, W, CV_8UC1);
	cvInitMatHeader(saturation, H, W, CV_8UC1);
*/
	//else
		return mask;
}


IplImage *Bumblebee2Control::IRgBy2(IplImage *image, IplImage *Irgby, IplImage *mask, CvPoint range0, CvPoint range1)
{
	double r, g, b;
	int SCALE = 0;
	int i = 0, j = 0;
	unsigned char p[3];

	int W, H;
	//対象人物領域の縦と横の長さを計算 range0 < range1
	//rang0のxあるいはyが0でない場合は1をたさなければならない

	if(range0.x != 0)W = range1.x - range0.x + 1;
	if(range0.y != 0)H = range1.y - range0.y + 1;
	if(range0.x == 0)W = abs(range1.x - range0.x);
	if(range0.y == 0)H = abs(range1.y - range0.y);

	//W = abs(range1.x - range0.x)+1;
	//H = abs(range1.y - range0.y)+1;
	if(range0.x == 0 && range0.y == 0 && range1.x == 0 && range0.y == 0)
		return NULL;
	if(W == 0)
		return NULL;

	double I_pre, Rg_pre, By_pre, hue_pre;

	//CvMat *I = NULL, *Rg = NULL, *By = NULL, *Rg_med = NULL, *By_med = NULL, *hue = NULL;
	//CvMat *I_filt = NULL, *MAD = NULL, *MAD_med = NULL, *saturation = NULL;
	
	//CvMat *I, *Rg, *By, *Rg_med, *By_med, *hue;
	//CvMat *I_filt, *MAD, *MAD_med, *saturation;
	int Q = W * H;
	double I_arr[Q], Rg_arr[Q], By_arr[Q], Rgm_arr[Q], Bym_arr[Q], hue_arr[Q];
	double If_arr[Q], MAD_arr[Q], MADm_arr[Q], sat_arr[Q];
/*
	double *I_arr, *Rg_arr, *By_arr, *Rgm_arr, *Bym_arr, *hue_arr;
	double *If_arr, *MAD_arr, *MADm_arr, *sat_arr;

	I_arr = new double[Q];
	Rg_arr = new double[Q];
	By_arr = new double[Q];
	Rgm_arr = new double[Q];
	Bym_arr = new double[Q];
	hue_arr = new double[Q];
	If_arr = new double[Q];
	MAD_arr = new double[Q];
	MADm_arr = new double[Q];
	sat_arr = new double[Q];
*/
	CvMat I, Rg, By, Rg_med, By_med, hue;
	CvMat I_filt, MAD, MAD_med, saturation;

	cvInitMatHeader(&I, H, W, CV_8UC1, I_arr);
	cvInitMatHeader(&Rg, H, W, CV_8UC1, Rg_arr);
	cvInitMatHeader(&By, H, W, CV_8UC1, By_arr);
	cvInitMatHeader(&Rg_med, H, W, CV_8UC1, Rgm_arr);
	cvInitMatHeader(&By_med, H, W, CV_8UC1, Bym_arr);
	cvInitMatHeader(&I_filt, H, W, CV_8UC1, If_arr);
	cvInitMatHeader(&hue, H, W, CV_8UC1, hue_arr);
	cvInitMatHeader(&MAD, H, W, CV_8UC1, MAD_arr);
	cvInitMatHeader(&MAD_med, H, W, CV_8UC1, MADm_arr);
	cvInitMatHeader(&saturation, H, W, CV_8UC1, sat_arr);

/*
	I = cvCreateMat(H, W, CV_8UC1);
	Rg = cvCreateMat(H, W, CV_8UC1);
	By = cvCreateMat(H, W, CV_8UC1);
	//CvMat *I_med = cvCreateMat(Hight, Width, CV_8UC1);
	Rg_med = cvCreateMat(H, W, CV_8UC1);
	By_med = cvCreateMat(H, W, CV_8UC1);
	hue = cvCreateMat(H, W, CV_8UC1);
	I_filt = cvCreateMat(H, W, CV_8UC1);
	MAD = cvCreateMat(H, W, CV_8UC1);
	MAD_med = cvCreateMat(H, W, CV_8UC1);
	saturation = cvCreateMat(H, W, CV_8UC1);
*/

	SCALE = (int)((W + H) / 320);		//メディアンフィルタのためのスケール計算 W != 0
	if(SCALE < 1 && SCALE <= 0)SCALE = 1;
	if(SCALE % 2 == 0)SCALE = SCALE + 1;	//OpenCVのメディアンフィルタ関数は偶数ではいけない

	int k = 0;

	//cout<<SCALE<<endl;
	//ピクセルデータ（R,G,B）を順次取得し，変換する
	for(int y = range0.y; y <= range1.y; y++){
		i = 0;
		for(int x = range0.x; x <= range1.x; x++){
			p[0] = image->imageData[image->widthStep * y + x * 3];        // B
			p[1] = image->imageData[image->widthStep * y + x * 3 + 1];    // G
			p[2] = image->imageData[image->widthStep * y + x * 3 + 2];    // R

			if(p[0] != 0 && p[1] != 0 && p[2] != 0){
				Logopp(p[0], b);
				Logopp(p[1], g);
				Logopp(p[2], r);

				I_pre   = (r + g + b) / 3.0;
				Rg_pre  = r - g;
				By_pre  = b - (g + r) / 2.0;
				hue_pre = atan2(Rg_pre, By_pre) * (180.0 / M_PI);

				I_arr[k] = I_pre;
				Rg_arr[k] = Rg_pre;
				By_arr[k] = By_pre;
				hue_arr[k] = hue_pre;
				k++; 
				//printf("%f\n",Rg_pre);
				//cvSetReal2D(I, j, i, I_pre);
				//cvSetReal2D(Rg, j, i, Rg_pre);
				//cvSetReal2D(By, j, i, By_pre);
				//cvSetReal2D(hue, j, i, hue_pre);
				//printf("%d %d %d %d\n", x, y, i, j);
				i++;
			}
			else{
				i++; 
				//I_arr[k] = 0.0;
				//Rg_arr[k] = 0.0;
				//By_arr[k] = 0.0;
				//hue_arr[k] = 0.0;				
				k++;
			}
		}
		j++;
	}
	//printf("%d %d\n", i, j);

	//自作メディアンフィルタ関数
	//MedianFilter(Rg, Rg_med, WIDTH, HEIGHT, 2 * 2 * SCALE);
	//MedianFilter(By, By_med, WIDTH, HEIGHT, 2 * 2 * SCALE);
	//メディアンフィルタ処理を行うときその範囲は1辺が3*SCALEの正方形となる.ただし,人物領域のサイズを越えてはいけない
	if(3 * 3 * SCALE * SCALE < W * H && 3 * SCALE < W && 3 * SCALE < H){
		cvSmooth(&Rg, &Rg_med, CV_MEDIAN, 3 * SCALE, 0, 0, 0);	//メディアンフィルタ
		cvSmooth(&By, &By_med, CV_MEDIAN, 3 * SCALE, 0, 0, 0);
	}
	else{
		cvSmooth(&Rg, &Rg_med, CV_MEDIAN, SCALE, 0, 0, 0);	//メディアンフィルタ
		cvSmooth(&By, &By_med, CV_MEDIAN, SCALE, 0, 0, 0);
	}

	//Compute texture amplitude
	//MedianFilter(I, I_filt, WIDTH, HEIGHT, 2 * 4 * SCALE);
	if(7 * 7 * SCALE * SCALE < W * H && 7 * SCALE < W && 7 * SCALE < H){
		cvSmooth(&I, &I_filt, CV_MEDIAN, 7 * SCALE, 0, 0, 0);	
	}
	else{
		cvSmooth(&I, &I_filt, CV_MEDIAN, SCALE, 0, 0, 0);
	}

	//hue,saturationを求める
	//printf("here\n");
	k = 0;

	for(int y = 0; y < j; y++){
		for(int x = 0; x < i; x++){
			//double Rg_data = cvGetReal2D(Rg_med, y, x);
			//double By_data = cvGetReal2D(By_med, y, x);
			//printf("%f\n", Rg_data);
			MAD_arr[k] = fabs(I_arr[k] - If_arr[k]);
			sat_arr[k] = sqrt(Rgm_arr[k] * Rgm_arr[k] + Bym_arr[k] * Bym_arr[k]);
			k++;
			//cvSetReal2D(MAD, y, x, cvGetReal2D(I, y, x) - cvGetReal2D(I_filt, y, x));
			//cvSetReal2D(MAD, y, x, fabs(cvGetReal2D(MAD, y, x)));
			//cvSetReal2D(saturation, y, x, sqrt(Rg_data * Rg_data + By_data * By_data));
			//cvSetReal2D(saturation, y, x, cvSqrt(cvGetReal2D(saturation, y, x)));			
		}
	}

 	//MAD_med = cvCloneImage(MAD);
	//MedianFilter(MAD, MAD_med, WIDTH, HEIGHT, 2 * 6 * SCALE);
	if(11 * 11 * SCALE * SCALE < W * H && 11 * SCALE < W && 11 * SCALE < H){
		cvSmooth(&MAD, &MAD_med, CV_MEDIAN, 11 * SCALE, 0, 0, 0);
	}
	else{
		cvSmooth(&MAD, &MAD_med, CV_MEDIAN, SCALE, 0, 0, 0);
	}

	i = 0;
	j = 0;
	k = 0;

	//マスク画像の作成
	cvSetZero(mask);	//ピクセル値をすべて0にする

	for(int y = range0.y; y <= range1.y; y++){
		i = 0;
		for(int x = range0.x; x <= range1.x; x++){
			//double MAD_data = cvGetReal2D(MAD_med, j, i);
			//double hue_data = cvGetReal2D(hue, j, i);
			//double saturation_data = cvGetReal2D(saturation, j, i);
			
			if(MADm_arr[k] < 4.5 && 120.0 < hue_arr[k] && hue_arr[k] < 160.0 && 10.0 < sat_arr[k] && sat_arr[k] < 60.0){
				mask->imageData[mask->widthStep * y + x] = (unsigned char)1;
			}

			else if(MADm_arr[k] < 4.5 && 150.0 < hue_arr[k] && hue_arr[k] < 180.0 && 20.0 < sat_arr[k] && sat_arr[k] < 80.0){
				mask->imageData[mask->widthStep * y + x] = (unsigned char)1;

			}
			i++;
			k++;
		}
		j++;
	}

	cvDilate(mask, mask, NULL, 5);	//膨張処理

	i = 0;
	j = 0;
	k = 0;
	//マスク画像の作成
	for(int y = range0.y; y <= range1.y; y++){
		i = 0;
		for(int x = range0.x; x <= range1.x; x++){
			//double hue_data = cvGetReal2D(hue, j, i);
			//double saturation_data = cvGetReal2D(saturation, j, i);
			if((int)(mask->imageData[mask->widthStep * y + x]) == 1 && 110.0 <= hue_arr[k] && hue_arr[k] <= 180.0 && 0.0 <= sat_arr[k] && sat_arr[k] <= 130.0){
				mask->imageData[mask->widthStep * y + x] = (unsigned char)1;//printf("clear\n");
			}
			else mask->imageData[mask->widthStep * y + x] = (unsigned char)0;
			i++;
			k++;
		}
		j++;
	}

	cvSetZero(Irgby);
	cvCopy(image, Irgby, mask);

	return mask;
}


double Bumblebee2Control::Logopp(unsigned char c, double &L)
{
	//double L = 0;
	L = 105.0 * log10((double)c + 1.0);
	return L;
}


double Bumblebee2Control::HeadDetect(CvPoint range0, CvPoint range1, float &pxx, float &pyy, float &pzz, float &detect_pz, int &min_x, int &min_y, int &detect_x, int &detect_y)
{
	int l = 0;
	//int min_x = 0, min_y = 0;
	//float pxx = 0.0, pyy = 0.0, pzz = 0.0;								//3次元座標
	//float detect_pz = 0.0;												//矩形域内で最大のZスピーカアレイ座標
	static float range_pz[WIDTH * HEIGHT];									//矩形域内のZスピーカアレイ座標
	static int range_x[WIDTH * HEIGHT], range_y[WIDTH * HEIGHT];			//矩形域内のX,Y画像座標			
	//int detect_x = 0, detect_y = 0;										//矩形域内で最大のZスピーカアレイ座標における画像座標X,Y
	//char buf[2048];		//add
	//printf("%d %d\n",xy1.x, xy.x);

	for(min_y = range0.y; min_y <= range0.y + 2; min_y++){
		for(min_x = range0.x; min_x < range1.x; min_x++){
  			Coordinate(min_x, min_y, pxx, pyy, pzz);//printf("%f\n", pzz);
			range_pz[l] = pzz;
			range_x[l] = min_x;
			range_y[l] = min_y;				
			//printf("%f %d %d\n", range_pz[l], range_x[l], l);				
			l++;
		}
	}	
	//printf("%d\n",l);

	for(int m = 0; m <= l; m++){
		if(range_pz[m] >= detect_pz){
			detect_pz = range_pz[m];
			detect_x = range_x[m];
			detect_y = range_y[m];
		}
	}
	return 0;
}


double Bumblebee2Control::HeadDetect2(IplImage *image, CvPoint range0, CvPoint range1, CvPoint &xy_min, CvPoint &xy_max,float &pxx, float &pyy, float &pzz, int &img_x, int &img_y, short *label)
{
	float px;
	float py;
	float pz;
	float keep_pz = 0.0;
	short keep_label = 0;
	
	CvPoint keep_min, keep_max;
							
	//pxx = 0.0;
	//pyy = 0.0;
	//pzz = 0.0;
	//img_x = 0;
	//img_y = 0;

	keep_min.x = keep_min.y = 1000;
	keep_max.x = keep_max.y = 0;

	if(range0.x > range1.x || range0.y > range1.y)
		return 0;

	for(int y = range0.y; y < range1.y; y++){
		for(int x = range0.x; x < range1.x; x++){
			if(0 < label[image->widthStep * y + x] && label[image->widthStep * y + x] < 4){
				Coordinate(x, y, px, py, pz);
				if(keep_pz < pz){
					//pxx = px;
					//pyy = py;
					keep_pz = pz;
					//img_x = x;
					//img_y = y;
					keep_label = label[image->widthStep * y + x];
				}
			}
		}
	}

	ofstream ofs;
	ofs.open( "keep_label.txt", ios::out | ios::app );

	ofs << keep_label << endl;

	ofs.close();

	for(int y = range0.y; y < range1.y; y++){
		for(int x = range0.x; x < range1.x; x++){
			if(label[image->widthStep * y + x] == keep_label){
				if(keep_min.x > x)
					keep_min.x = x;
				if(keep_min.y > y)
					keep_min.y = y;
				if(keep_max.x < x)
					keep_max.x = x;
				if(keep_min.y < y)
					keep_max.y = y;
			}
		}
	}

//if(keep_min.x >= 0 && keep_min.y >= 0 && keep_max.x < WIDTH && keep_max.y < HEIGHT){
	xy_min.x = keep_min.x;
	xy_min.y = keep_min.y;
	xy_max.x = keep_max.x;	
	xy_max.y = keep_max.y;	
//}
	img_x = (int)((xy_max.x - xy_min.x) / 2 + 1 + xy_min.x);
	img_y = (int)((xy_max.y - xy_min.y) / 2 + 1 + xy_min.y);
	Coordinate(img_x, img_y, pxx, pyy, pzz);

	return 0;
}


IplImage *Bumblebee2Control::Labeling( IplImage *img, IplImage *_mask, short *label )
{
/*
    CvFont dfont;
    float hscale      = 0.5f;
    float vscale      = 0.5f;
    float italicscale = 0.0f;
    int  thickness    = 1;
*/
	//short *label;


    //if( !img ) 						//画像が取得されていなければ,処理を行わない
        //return 0;

	//ラベリング処理の結果保存用配列
	//label = new short[ _mask->width * _mask->height ];

    //ラベリング実行
    labeling.Exec((unsigned char *)_mask->imageData, label, _mask->width, _mask->height, true, 100);
    //------------------------------------------------------------------------------
    //int Exec( SrcT *target, DstT *result, int target_width, int target_height,
    //          const bool is_sort_region, const int region_size_min )
    //引数:
    //  SrcT *target 入力バッファ
    //  DstT *result 出力バッファ
    //  int target_width バッファの横サイズ
    //  int target_height バッファの縦サイズ
    //  bool is_sort_region 連続領域を大きさの降順にソートするか(1番領域の大きいサイズは１になるはず)
    //  int region_size_min 最小の領域サイズ(これ未満は無視する)
    //------------------------------------------------------------------------------

	//領域数を出力
	//cout << "number of regions: " << labeling.GetNumOfRegions() << endl;

	for(int i = 0; i < _mask->width * _mask->height; i++){
		if(0 < (int)label[i] && (int)label[i] < 4){
			_mask->imageData[i] = 1;
		}
		else _mask->imageData[i] = 0;
	}
/*
	if(skin != NULL){
		memset(skin->imageData, 0, WIDTH * HEIGHT * 3);
		cvCopy(img, skin, _mask);

		char text[255], str[255];
		int num = labeling.GetNumOfRegions();
		strcpy( str, "number of regions: " );
		sprintf( text, "%s %d", str, num );
		cvInitFont (&dfont, CV_FONT_HERSHEY_SIMPLEX , hscale, vscale, italicscale, thickness, CV_AA);
		cvPutText(skin, text, cvPoint(10, 15), &dfont, CV_RGB(255, 255, 255));

		cvNamedWindow("Skin", 1);
		cvShowImage("Skin", skin);
	}
*/
	return _mask;
}


double Bumblebee2Control::LabelDetect( IplImage *mask_i, short *label, CvPoint p_center )
{
	track_object  = 0;
	track_object2 = 0;
	track_object3 = 0;

	//ラベリングされた領域をそれぞれ矩形で囲む
	//ラベリングされた各領域での画像座標を求める
	int p1[2], p2[2], p3[2];
	CvPoint point1_max, point2_max, point3_max;
	CvPoint point1_min, point2_min, point3_min;
//if(key == 'g')	fixed_hist = !fixed_hist;
//if(fixed_hist){
	point1_max.x = point2_max.x = point3_max.x = 0;
	point1_max.y = point2_max.y = point3_max.y = 0;
	point1_min.x = point2_min.x = point3_min.x = 1000;
	point1_min.y = point2_min.y = point3_min.y = 1000;

short label_center;
label_center = label[mask_i->widthStep * p_center.y + p_center.x];
	for(int y = 0; y < mask_i->height; y++){
		for(int x = 0; x < mask_i->width; x++){

			if(p_center.x != 0 && p_center.y !=0){				
				if(label[mask_i->widthStep * y + x] == label_center && label_center == 1){
					p1[0] = x;
					p1[1] = y;
					if(p1[0] >= point1_max.x){			//領域1におけるx座標の最大値とそのときのy座標を求める
						point1_max.x = p1[0];
					}
					if(p1[1] >= point1_max.y){			//領域1におけるy座標の最大値とそのときのx座標を求める
						point1_max.y = p1[1];
					}
					if(p1[0] <= point1_min.x){			//領域1におけるx座標の最小値とそのときのy座標を求める
						point1_min.x = p1[0];
					}
					if(p1[1] <= point1_min.y){			//領域1におけるy座標の最小値とそのときのx座標を求める
						point1_min.y = p1[1];
					}
				}
			}
			else{
				if(label[mask_i->widthStep * y + x] == 1){
					p1[0] = x;
					p1[1] = y;
					if(p1[0] >= point1_max.x){			//領域1におけるx座標の最大値とそのときのy座標を求める
						point1_max.x = p1[0];
					}
					if(p1[1] >= point1_max.y){			//領域1におけるy座標の最大値とそのときのx座標を求める
						point1_max.y = p1[1];
					}
					if(p1[0] <= point1_min.x){			//領域1におけるx座標の最小値とそのときのy座標を求める
						point1_min.x = p1[0];
					}
					if(p1[1] <= point1_min.y){			//領域1におけるy座標の最小値とそのときのx座標を求める
						point1_min.y = p1[1];
					}
				}
			}

			if(label[mask_i->widthStep * y + x] == 2){
				p2[0] = x;
				p2[1] = y;
				if(p2[0] >= point2_max.x){			//領域2におけるx座標の最大値とそのときのy座標を求める
					point2_max.x = p2[0];
				}
				if(p2[1] >= point2_max.y){			//領域2におけるy座標の最大値とそのときのx座標を求める
					point2_max.y = p2[1];
				}
				if(p2[0] <= point2_min.x){			//領域2におけるx座標の最小値とそのときのy座標を求める
					point2_min.x = p2[0];
				}
				if(p2[1] <= point2_min.y){			//領域2におけるy座標の最小値とそのときのx座標を求める
					point2_min.y = p2[1];
				}
			}

			if(label[mask_i->widthStep * y + x] == 3){
				p3[0] = x;
				p3[1] = y;
				if(p3[0] >= point3_max.x){			//領域3におけるx座標の最大値とそのときのy座標を求める
					point3_max.x = p3[0];
				}
				if(p3[1] >= point3_max.y){			//領域3におけるy座標の最大値とそのときのx座標を求める
					point3_max.y = p3[1];
				}
				if(p3[0] <= point3_min.x){			//領域3におけるx座標の最小値とそのときのy座標を求める
					point3_min.x = p3[0];
				}
				if(p3[1] <= point3_min.y){			//領域3におけるy座標の最小値とそのときのx座標を求める
					point3_min.y = p3[1];
				}
			}
		}
	}

	int w1, w2, w3, h1, h2, h3;
	w1 = point1_max.x - point1_min.x;
	h1 = point1_max.y - point1_min.y;
	w2 = point2_max.x - point2_min.x;
	h2 = point2_max.y - point2_min.y;
	w3 = point3_max.x - point3_min.x;
	h3 = point3_max.y - point3_min.y;


	//原点および選択された領域を設定

	origin  = point1_min;
	origin2 = point2_min;
	origin3 = point3_min;

	selection  = cvRect( point1_min.x, point1_min.y, w1, h1 );
	selection2 = cvRect( point2_min.x, point2_min.y, w2, h2 );
	selection3 = cvRect( point3_min.x, point3_min.y, w3, h3 );
	//fixed_hist = !fixed_hist;

	//widthとheightがどちらも正であれば,track_objectフラグを開始フラグにする
	if( selection.width > 0 && selection.height > 0 ){
		track_object = -1;
		if( selection2.width > 0 && selection2.height > 0 ){
			track_object2 = -1;
			if( selection3.width > 0 && selection3.height > 0 ){
				track_object3 = -1;
			}
		}
	}
//}

	return 0;
}


CvScalar Bumblebee2Control::hsv2rgb( float hue_cs )
{
  //----------------------------------------------------------
  //	入力された1つの色相値をRGBに変換する
  //
  //	引数    :
  //	hue_cs	    : HSV表色系における色相値H
  //	戻り値  ：
  //	CvScalar: RGBの色情報がBGRの順で格納されたコンテナ
  //----------------------------------------------------------

    int rgb[3], p, sector;
    static const int sector_data[][3] =
        {{0,2,1}, {1,2,0}, {1,0,2}, {2,0,1}, {2,1,0}, {0,1,2}};
    hue_cs *= 0.033333333333333333333333333333333f;
    sector = cvFloor(hue_cs);
    p = cvRound(255*(hue_cs - sector));
    p ^= sector & 1 ? 255 : 0;

    rgb[sector_data[sector][0]] = 255;
    rgb[sector_data[sector][1]] = 0;
    rgb[sector_data[sector][2]] = p;

    return cvScalar(rgb[2], rgb[1], rgb[0], 0);
}


int Bumblebee2Control::InitCamShift()
{
	_theta = 55;
	fixed_hist = false;
	select_object  = 0;
	select_object2 = 0;
	select_object3 = 0;			//初期追跡領域の設定判別フラグ値 (0: 設定なし, 1: 設定あり)
	track_object  = 0;
	track_object2 = 0;
	track_object3 = 0;			//トラッキングの開始/停止用フラグ値 (0: 停止, -1: 開始, 1: トラッキング中)
	backproject_mode = 0;
	hdims = 256;					//ヒストグラムの次元数
	vmin = 10;					//明度の最大値
	vmax = 256;					//明度の最小値
	smin = 30;					//彩度の最小値
	hranges_arr[0] = 0;			//ヒストグラムのレンジ
	hranges_arr[1] = 180;		//ヒストグラムのレンジ
	hranges = hranges_arr;
	hist = 0;
	hist_h = 0;
	hist2 = 0;
	hist_h2 = 0;
	hsv = NULL;
	hue_cs = NULL;
	mask_cs = NULL;
	backproject = NULL;
	histimg = NULL;
	histimg2 = NULL;
	histimg_h = NULL;
	histimg_h2 = NULL;
	return 0;
}


void Bumblebee2Control::MultiCS( IplImage *camshift, IplImage *histimg_num, IplImage *histimg_h_num, CvHistogram *hist_num, int object_num, CvRect selection_num, CvBox2D &track_box_num, CvConnectedComp track_comp_num, short key, int num)
{

	int bin_w, bin_num;

	if(key == 'g'){
		fixed_hist = false;//fixed_hist = !fixed_hist;
	}

	if(object_num){
	  	int _vmin = vmin, _vmax = vmax;

		cvInRangeS( hsv, cvScalar(0, smin, MIN(_vmin, _vmax), 0),
		            cvScalar(180, 256, MAX(_vmin, _vmax), 0), mask_cs );
		cvSplit( hsv, hue_cs, 0, 0, 0 );

        if( object_num < 0 )
        {
            float max_val = 0.f;
			cvSetImageROI( hue_cs, selection_num );
            cvSetImageROI( mask_cs, selection_num );
            cvCalcHist( &hue_cs, hist_num, 0, mask_cs );	//ヒストグラムを計算し,最大値を求める
            cvGetMinMaxHistValue( hist_num, 0, &max_val, 0, 0 );
			//ヒストグラムの縦軸（頻度）を0-255のダイナミックレンジに正規化
            cvConvertScale( hist_num->bins, hist_num->bins, max_val ? 255. / max_val : 0., 0 );	
            cvResetImageROI( hue_cs );				//hue_cs,mask画像に設定されたROIをリセット
            cvResetImageROI( mask_cs );
            track_window = selection_num;
            object_num = 1;					//track_objectをトラッキング中にする

            cvZero( histimg_num );
            bin_w = histimg_num->width / hdims;
            for(int i = 0; i < hdims; i++ ){
                int val = cvRound( cvGetReal1D(hist_num->bins, i) * histimg_num->height / 255 );
                CvScalar color = hsv2rgb(i * 180.f / hdims);
                cvRectangle( histimg_num, cvPoint(i * bin_w, histimg_num->height),
                             cvPoint((i + 1) * bin_w, histimg_num->height - val),
                             color, -1, 8, 0 );
            }
			if(num == 1){
				cvNamedWindow("Histogram 1", CV_WINDOW_AUTOSIZE);
				cvShowImage( "Histogram 1", histimg_num );
			}
			if(num == 2){
				cvNamedWindow("Histogram 2", CV_WINDOW_AUTOSIZE);
				cvShowImage( "Histogram 2", histimg_num );
			}

			if(key == 'p'){
				if(num == 1){
					hist_h = hist_num;
					track_window_h = selection_num;					
				}
				if(num == 2){
					hist_h2 = hist_num;
					track_window_h2 = selection_num;					
				}
				fixed_hist = true;
			}

			if(fixed_hist){				
				cvZero( histimg_h_num );
			    bin_num = histimg_h_num->width / hdims;
				if(num == 1){
					for(int i = 0; i < hdims; i++ ){
					    int val = cvRound( cvGetReal1D(hist_h->bins, i) * histimg_h_num->height / 255 );
					    CvScalar color = hsv2rgb(i * 180.f / hdims);
					    cvRectangle( histimg_h_num, cvPoint(i * bin_num, histimg_h_num->height),
					                 cvPoint((i + 1) * bin_num, histimg_h_num->height - val),
					                 color, -1, 8, 0 );
					}
					cvNamedWindow("Fixed Histogram 1", CV_WINDOW_AUTOSIZE);
					cvShowImage( "Fixed Histogram 1", histimg_h_num );
				}
				if(num == 2){
					for(int i = 0; i < hdims; i++ ){
					    int val = cvRound( cvGetReal1D(hist_h2->bins, i) * histimg_h_num->height / 255 );
					    CvScalar color = hsv2rgb(i * 180.f / hdims);
					    cvRectangle( histimg_h_num, cvPoint(i * bin_num, histimg_h_num->height),
					                 cvPoint((i + 1) * bin_num, histimg_h_num->height - val),
					                 color, -1, 8, 0 );
					}
					cvNamedWindow("Fixed Histogram 2", CV_WINDOW_AUTOSIZE);
					cvShowImage( "Fixed Histogram 2", histimg_h_num );
				}
			}
				
			if(!fixed_hist){
			    cvCalcBackProject( &hue_cs, backproject, hist_num );	//バックプロジェクションを計算する
			    cvAnd( backproject, mask_cs, backproject, 0 );		//backProjectionのうち,マスクが1であるとされた部分のみ残す
			    cvCamShift( backproject, track_window,					//CamShift法による領域追跡を実行する
			                cvTermCriteria( CV_TERMCRIT_EPS | CV_TERMCRIT_ITER, 10, 1 ),
			                &track_comp_num, &track_box_num );
			    track_window = track_comp_num.rect;
			    
			    if( backproject_mode )
			        cvCvtColor( backproject, camshift, CV_GRAY2BGR );
			    if( !camshift->origin )
			        track_box_num.angle = -track_box_num.angle;

				if(num == 1)
			    	cvEllipseBox( camshift, track_box_num, CV_RGB(255, 0, 0), 2, CV_AA, 0 );
				if(num == 2)
				    cvEllipseBox( camshift, track_box_num, CV_RGB(0, 0, 255), 2, CV_AA, 0 );
			}
      	}
	}

	if(fixed_hist){
		if(num == 1){
			cvCalcBackProject( &hue_cs, backproject, hist_h );	//バックプロジェクションを計算する
			cvAnd( backproject, mask_cs, backproject, 0 );		//backProjectionのうち,マスクが1であるとされた部分のみ残す
			cvCamShift( backproject, track_window_h,			//CamShift法による領域追跡を実行する
				        cvTermCriteria( CV_TERMCRIT_EPS | CV_TERMCRIT_ITER, 10, 1 ),
				        &track_comp_num, &track_box_num );
			track_window_h = track_comp_num.rect;
		}
		if(num == 2){
			cvCalcBackProject( &hue_cs, backproject, hist_h2 );	//バックプロジェクションを計算する
			cvAnd( backproject, mask_cs, backproject, 0 );		//backProjectionのうち,マスクが1であるとされた部分のみ残す
			cvCamShift( backproject, track_window_h2,			//CamShift法による領域追跡を実行する
				        cvTermCriteria( CV_TERMCRIT_EPS | CV_TERMCRIT_ITER, 10, 1 ),
				        &track_comp_num, &track_box_num );
			track_window_h2 = track_comp_num.rect;
		}

		if( backproject_mode )
		    cvCvtColor( backproject, camshift, CV_GRAY2BGR );
		if( !camshift->origin )
		    track_box_num.angle = -track_box_num.angle;

		if(track_box_num.size.width == 0 && track_box_num.size.height == 0){
			//track_window_h = selection;
			//printf("lost\n");
		}
		if(num == 1)
			cvEllipseBox( camshift, track_box_num, CV_RGB(255, 0, 0), 2, CV_AA, 0 );
		if(num == 2)
			cvEllipseBox( camshift, track_box_num, CV_RGB(0, 0, 255), 2, CV_AA, 0 );
	}
}


IplImage *Bumblebee2Control::CamShift(IplImage *image, IplImage *camshift, IplImage *camshift2, CvBox2D &box1, CvBox2D &box2, short key)
{
    //cvNamedWindow( "Histogram", CV_WINDOW_AUTOSIZE );
    cvNamedWindow( "CamShiftDemo", CV_WINDOW_AUTOSIZE );
    //cvSetMouseCallback( "CamShiftDemo", on_mouse, 0 );
    cvCreateTrackbar( "Vmin", "CamShiftDemo", &vmin, 256, 0 );
    cvCreateTrackbar( "Vmax", "CamShiftDemo", &vmax, 256, 0 );
    cvCreateTrackbar( "Smin", "CamShiftDemo", &smin, 256, 0 );
	
    cvCreateTrackbar( "Theta", "CamShiftDemo", &_theta, 180, 0 );

    if( !camshift )
    {
		//hist = 0;
        //allocate all the buffers
        camshift = cvCreateImage( cvGetSize(image), IPL_DEPTH_8U, 3 );
        camshift->origin = image->origin;
    
        hsv = cvCreateImage( cvGetSize(image), IPL_DEPTH_8U, 3 );
        hue_cs = cvCreateImage( cvGetSize(image), IPL_DEPTH_8U, 1 );
        mask_cs = cvCreateImage( cvGetSize(image), IPL_DEPTH_8U, 1 );
        backproject = cvCreateImage( cvGetSize(image), IPL_DEPTH_8U, 1 );
        hist = cvCreateHist( 1, &hdims, CV_HIST_ARRAY, &hranges, 1 );
        //hist_h = cvCreateHist( 1, &hdims, CV_HIST_ARRAY, &hranges, 1 );
        hist2 = cvCreateHist( 1, &hdims, CV_HIST_ARRAY, &hranges, 1 );
        //hist_h2 = cvCreateHist( 1, &hdims, CV_HIST_ARRAY, &hranges, 1 );
        histimg = cvCreateImage( cvSize(320, 200), IPL_DEPTH_8U, 3 );
        histimg2 = cvCreateImage( cvSize(320, 200), IPL_DEPTH_8U, 3 );
        histimg_h = cvCreateImage( cvSize(320, 200), IPL_DEPTH_8U, 3 );
        histimg_h2 = cvCreateImage( cvSize(320, 200), IPL_DEPTH_8U, 3 );
        cvZero( histimg );
        cvZero( histimg2 );
	}
    cvCopy( image, camshift, 0 );
    cvCvtColor( camshift, hsv, CV_BGR2HSV );		    //キャプチャされた画像をHSV表色系に変換してhsvに格納

	//printf("%d\n", track_object);
	//track_objectフラグが0以下なら,以下の処理を行う
	MultiCS(camshift, histimg, histimg_h, hist, track_object, selection, track_box, track_comp, key, 1);
	//MultiCS(camshift, histimg2, histimg_h2, hist2, track_object2, selection2, track_box2, track_comp2, key, 2);
	box1 = track_box;
	box2 = track_box2;

	cvCopy( camshift, camshift2, 0 );
    cvShowImage( "CamShiftDemo", camshift );
   // cvShowImage( "Histogram", histimg );


        switch( (char) key )
        {
		    case 'b':
		        backproject_mode ^= 1;
		        break;
		    case 'c':
		        track_object  = 0;
				track_object2 = 0;
				track_object3 = 0;
		        cvZero( histimg );
		        break;
/*		    case 'h':
		        show_hist ^= 1;
		        if( !show_hist )
		            cvDestroyWindow( "Histogram" );
		        else
		            cvNamedWindow( "Histogram", 1 );
		        break;
*/		    default:
		        ;
        }

    //}

    //cvReleaseCapture( &capture );
    //cvDestroyWindow("CamShiftDemo");

	return camshift;
}


